#------------------------------------------------------------------------------
# Import Libraries
#------------------------------------------------------------------------------
import bpy
import bpy_extras
import os
import sys
import math
import numpy as np
from math import radians, sqrt, degrees
from mathutils import Matrix
from mathutils import Vector
from time import sleep, time
# from visual_fire_power import eulerAnglesToRotationMatrix

#------------------------------------------------------------------------------
# Borrowed Functions
#------------------------------------------------------------------------------
def round_to_zero(array, threshold=1e-6):
    """Return given array with all values below threshold set to zero"""
    return np.where(np.abs(array)>threshold, array, 0)

def eulerAnglesToRotationMatrix(theta):
    """ Convert list of euler angles [X, Y, Z] following the Tait-Bryan
    to a 3x3 rotation matrix, but with a flip in coordinates required
    to match OpenCV.

    Edited from https://learnopencv.com/rotation-matrix-to-euler-angles/"""

    R_x = np.array([[1,         0,                  0                   ],
                    [0,         math.cos(theta[0]), math.sin(theta[0]) ],
                    [0,         -math.sin(theta[0]), math.cos(theta[0])  ]
                    ])

    R_y = np.array([[math.cos(theta[1]),    0,      -math.sin(theta[1])  ],
                    [0,                     1,      0                   ],
                    [math.sin(theta[1]),   0,      math.cos(theta[1])  ]
                    ])

    R_z = np.array([[math.cos(theta[2]),    math.sin(theta[2]),    0],
                    [-math.sin(theta[2]),    math.cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])

    R = np.dot(R_x, np.dot(R_y, R_z))
    # Flip required to match blender angles
    cv_flip = np.array([[1, 0,  0],
                        [0,-1,  0],
                        [0, 0, -1]])
    R = np.dot(cv_flip, R)
    R = round_to_zero(R)

    return R

def is_face_skip(f):
    """Ignore faces that pass this test"""
    return f.hide is False
    # You may want to filter based on material.
    # return f.material_index == 0


def bmesh_from_object_final(ob):
    """Return a bmesh version of a normal blender mesh"""
    import bmesh
    USE_FILTER_FACES = True
    matrix = ob.matrix_world
    me = ob.to_mesh()
    me.transform(matrix)
    bm = bmesh.new()
    bm.from_mesh(me)
    # bpy.data.meshes.remove(me)
    if USE_FILTER_FACES:
        faces_remove = [f for f in bm.faces if not is_face_skip(f)]
        for f in faces_remove:
            bm.faces.remove(f)
    return (bm, matrix.is_negative)


def volume_and_area_from_object(ob):
    """ Return volume and surface area of blender mesh object

    See notes on this at:
    https://blender.stackexchange.com/questions/99672
    /how-to-get-the-final-volume-area-of-all-selected-objects-ignoring-some-faces
    """
    bm, is_negative = bmesh_from_object_final(ob)
    volume = bm.calc_volume(signed=True)
    area = sum(f.calc_area() for f in bm.faces)
    bm.free()
    volume = abs(volume)
    return volume, area


def get_calibration_matrix_K_from_blender(camd, resolution=None):
    """Build intrinsic camera parameters from Blender camera data

    See notes on this at:
    blender.stackexchange.com/questions/15102
    /what-is-blenders-camera-projection-matrix-model
    """
    f_in_mm = camd.lens
    scene = bpy.context.scene
    resolution_x_in_px = resolution[0]
    resolution_y_in_px = resolution[1]
    scale = scene.render.resolution_percentage / 100
    sensor_width_in_mm = camd.sensor_width
    sensor_height_in_mm = camd.sensor_height
    pixel_aspect_ratio = scene.render.pixel_aspect_x / scene.render.pixel_aspect_y
    if (camd.sensor_fit == 'VERTICAL'):
        # the sensor height is fixed (sensor fit is horizontal),
        # the sensor width is effectively changed with the pixel aspect ratio
        s_u = resolution_x_in_px * scale / sensor_width_in_mm / pixel_aspect_ratio
        s_v = resolution_y_in_px * scale / sensor_height_in_mm
    else: # 'HORIZONTAL' and 'AUTO'
        # the sensor width is fixed (sensor fit is horizontal),
        # the sensor height is effectively changed with the pixel aspect ratio
        pixel_aspect_ratio = scene.render.pixel_aspect_x / scene.render.pixel_aspect_y
        s_u = resolution_x_in_px * scale / sensor_width_in_mm
        s_v = resolution_y_in_px * scale * pixel_aspect_ratio / sensor_height_in_mm

    # Parameters of intrinsic calibration matrix K
    alpha_u = f_in_mm * s_u
    alpha_v = f_in_mm * s_v
    u_0 = resolution_x_in_px * scale / 2
    v_0 = resolution_y_in_px * scale / 2
    skew = 0 # only use rectangular pixels

    K = Matrix(
        ((alpha_u,       0, u_0),
         (      0, alpha_u, v_0),
         (      0,       0, 1  ))
        )
    return K


def get_3x4_RT_matrix_from_blender(cam):
    """Return camera rotation and translation matrices from Blender.

     There are 3 coordinate systems involved:
        1. The World coordinates: "world"
           - right-handed
        2. The Blender camera coordinates: "bcam"
           - x is horizontal
           - y is up
           - right-handed: negative z look-at direction
        3. The desired computer vision camera coordinates: "cv"
           - x is horizontal
           - y is down (to align to the actual pixel coordinates
             used in digital images)
           - right-handed: positive z look-at direction"""

    # bcam stands for blender camera
    R_bcam2cv = Matrix(
        ((1, 0,  0),
         (0, 1, 0),
         (0, 0, -1)))

    # Transpose since the rotation is object rotation,
    # and we want coordinate rotation
    R_world2bcam = cam.rotation_euler.to_matrix().transposed()
    T_world2bcam = -1*R_world2bcam @ cam.location
    #
    # Use matrix_world instead to account for all constraints
    # location, rotation = cam.matrix_world.decompose()[0:2]
    # R_world2bcam = rotation.to_matrix().transposed()

    # Convert camera location to translation vector used in coordinate changes
    # T_world2bcam = -1*R_world2bcam*cam.location
    # Use location from matrix_world to account for constraints:
    #T_world2bcam = -1*R_world2bcam @ location

    # Build the coordinate transform matrix from world to computer vision camera
    # NOTE: Use * instead of @ here for older versions of Blender
    # TODO: detect Blender version
    R_world2cv = R_bcam2cv@R_world2bcam
    T_world2cv = R_bcam2cv@T_world2bcam

    # put into 3x4 matrix
    RT = Matrix((
        R_world2cv[0][:] + (T_world2cv[0],),
        R_world2cv[1][:] + (T_world2cv[1],),
        R_world2cv[2][:] + (T_world2cv[2],)
         ))
    return RT

def get_3x4_P_matrix_from_blender(cam, resolution=None):
    """Return projection matrix of a blender camera"""
    K = get_calibration_matrix_K_from_blender(cam.data, resolution)
    RT = get_3x4_RT_matrix_from_blender(cam)
    return K@RT, K, RT

#------------------------------------------------------------------------------
# Original Functions
#------------------------------------------------------------------------------
def create_camera(
        name, location=(0,0,0),
        rotation=(0,0,0), focal_length=50,
        sensor_width=36):
    """Create a camera object in blender"""

    scn = bpy.context.scene

    cam = bpy.data.cameras.new(name)
    cam.lens = focal_length
    cam.sensor_width = sensor_width

    cam_obj = bpy.data.objects.new(name, cam)
    cam_obj.location = location
    cam_obj.rotation_euler = rotation

    scn.collection.objects.link(cam_obj)

    return

def create_shape(shape="Cube", rotation=(0,0,0)):
    """Create primitive shape at the origin with colored material"""
    if shape == "Cube":
        bpy.ops.mesh.primitive_cube_add(rotation=rotation)
    elif shape == "Cone":
        bpy.ops.mesh.primitive_cone_add(rotation=rotation)
    elif shape == "Cylinder":
        bpy.ops.mesh.primitive_cylinder_add(rotation=rotation)
    elif shape == "Sphere":
        bpy.ops.mesh.primitive_uv_sphere_add(rotation=rotation)
    elif shape == "Iso-Sphere":
        bpy.ops.mesh.primitive_ico_sphere_add(rotation=rotation)
    elif shape == "Torus":
        bpy.ops.mesh.primitive_torus_add(rotation=rotation)

    obj = bpy.context.active_object

    # Get material
    mat = bpy.data.materials.get("Material")
    if mat is None:
        # create material
        mat = bpy.data.materials.new(name="Material")

    mat.node_tree.nodes["Principled BSDF"].inputs[0].default_value = (1,0,0,1)

    # Assign it to object
    if obj.data.materials:
        # assign to 1st material slot
        obj.data.materials[0] = mat
    else:
        # no slots
        obj.data.materials.append(mat)

    return

def render_all_cams(output_dir=None, resolution=(1920, 1080)):
    """Save a render from each camera in the scene"""
    scn = bpy.context.scene
    if output_dir is None:
        output_dir = os.getcwd()+"\\"
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    for obj in scn.objects:
        if obj.type == 'CAMERA':
            scn.camera = obj
            scn.render.filepath = os.path.join(output_dir,obj.name)
            scn.render.resolution_x = resolution[0]
            scn.render.resolution_y = resolution[1]
            bpy.ops.render.render(write_still=True)
            sleep(2)

    return

def render_from_cam(cam, output_dir=None, resolution=(1920, 1080)):
    """Save a render from each camera in the scene"""
    scn = bpy.context.scene
    if output_dir is None:
        output_dir = os.getcwd()+"\\"
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    scn.camera = cam
    scn.render.filepath = os.path.join(output_dir,cam.name)
    scn.render.resolution_x = resolution[0]
    scn.render.resolution_y = resolution[1]
    bpy.ops.render.render(write_still=True)
    sleep(2)
    bpy.ops.wm.window_close()

    return

def save_all_camera_params(output_dir, resolution, camera_location, camera_rotation):
    """ Save the projection matrix, camera matrix, and pose of each camera in the scene """
    scn = bpy.context.scene
    # if output_dir is None:
    #     output_dir = os.getcwd()
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    for obj in scn.objects:
        if obj.type == 'CAMERA':
            P, K, RT = get_3x4_P_matrix_from_blender(obj, resolution)
            RT = get_RT_manually(camera_location, camera_rotation)

            nP = np.matrix(P)
            nK = np.matrix(K)
            nRT = np.matrix(RT)

            np.save(os.path.join(output_dir, obj.name+"-K.npy"), nK)
            np.save(os.path.join(output_dir, obj.name+"-RT.npy"), nRT)

    return

def get_shape_properties(so):
    """ Return volume, area, and X, Y, Z dimensions of a blender shape """
    vol, area = volume_and_area_from_object(so)
    shape_x = so.dimensions.x
    shape_y = so.dimensions.y
    shape_z = so.dimensions.z
    shape_properties = np.array([vol,area,shape_x,shape_y,shape_z]).flatten()

    return shape_properties

def create_light():
    """ Create a light source in the scene """
    # create light datablock, set attributes
    light_data = bpy.data.lights.new(name="light_2.80", type='POINT')
    light_data.energy = 1000

    # create new object with our light datablock
    light_object = bpy.data.objects.new(name="light_2.80", object_data=light_data)

    # link light object
    bpy.context.collection.objects.link(light_object)

    # make it active
    bpy.context.view_layer.objects.active = light_object

    #change location
    light_object.location = (-5,-5,5)

    # update scene, if needed
    dg = bpy.context.evaluated_depsgraph_get()
    dg.update()

    return

def change_color(obj):
    """Change the color of a blender shape to red (for thresholding)"""
    for slot in bpy.data.objects[new_obj.name].material_slots:
        new_mat = bpy.data.materials.new(name="carpaint.NewRed")
        new_mat.diffuse_color = (1,0,0)
        slot.material = new_mat

    return

def get_RT_manually(cam_location, cam_rotation):
    """Return pose matrix from the [X, Y, Z] camera location and
    [theta_x, theta_y, theta_z] euler angles of camera rotation"""
    tvec = np.array(cam_location)
    R = eulerAnglesToRotationMatrix(np.array(cam_rotation))
    RT = np.hstack([R, tvec[:, np.newaxis]])
    return RT
