import numpy as np
import cv2 as cv
import math
from math import radians

def round_to_zero(array, threshold=1e-6):
    """Return given array with all values below threshold set to zero"""
    return np.where(np.abs(array)>threshold, array, 0)

def eulerAnglesToRotationMatrix(theta):
    """ Convert list of euler angles [X, Y, Z] following the Tait-Bryan
    to a 3x3 rotation matrix, but with a flip in coordinates required
    to match OpenCV.

    Edited from https://learnopencv.com/rotation-matrix-to-euler-angles/"""

    R_x = np.array([[1,         0,                  0                   ],
                    [0,         math.cos(theta[0]), math.sin(theta[0]) ],
                    [0,         -math.sin(theta[0]), math.cos(theta[0])  ]
                    ])

    R_y = np.array([[math.cos(theta[1]),    0,      -math.sin(theta[1])  ],
                    [0,                     1,      0                   ],
                    [math.sin(theta[1]),   0,      math.cos(theta[1])  ]
                    ])

    R_z = np.array([[math.cos(theta[2]),    math.sin(theta[2]),    0],
                    [-math.sin(theta[2]),    math.cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])

    R = np.dot(R_x, np.dot(R_y, R_z))
    # Flip required to match blender angles
    cv_flip = np.array([[1, 0,  0],
                        [0,-1,  0],
                        [0, 0, -1]])
    R = np.dot(cv_flip, R)
    R = round_to_zero(R)

    return R

def isRotationMatrix(R) :
    """Check if a numpy array is a valid rotation matrix"""
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype = R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6

def rotationMatrixToEulerAngles(R) :
    """Convert a 3x3 OpenCV rotation matrix to the Euler angles used by
    Blender"""

    assert(isRotationMatrix(R))

    sy = math.sqrt(R[0,0] * R[0,0] +  R[1,0] * R[1,0])

    singular = sy < 1e-6

    if  not singular :
        x = math.atan2(R[2,1] , R[2,2])
        z = math.atan2(-R[2,0], sy)
        y = math.atan2(R[1,0], R[0,0])
    else :
        x = math.atan2(-R[1,2], R[1,1])
        z = math.atan2(-R[2,0], sy)
        y = 0

    angles = np.array([x, y, z])
    angles = round_to_zero(angles)

    return np.array([x, y, z])

def inverse_camera_matrix(K):
    """Return inverted OpenCV intrinsic camera matrix"""
    f_x = K[0,0]
    f_y = K[1,1]
    c_x = K[0,2]
    c_y = K[1,2]
    alpha = 1/(f_x*f_y)

    K_inv = np.float32([
        [f_y,   0, -f_y * c_x],
        [  0, f_x, -f_x * c_y],
        [  0,   0,    f_x*f_y]
        ]) * alpha

    return K_inv

def get_perspective_matrix(
    K,
    RT,
    new_location=(-5*np.sqrt(2),-5*np.sqrt(2),-0.2),
    new_rotation=(90,0,-55)):
    """Return the homography that transforms an image from the current camera
    to one of the same scene taken by the same camera at a new location and
    rotation, assuming the same projective distance from the scene"""
    theta = [radians(angle) for angle in new_rotation]
    R_new = eulerAnglesToRotationMatrix(theta)

    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    K_inv = inverse_camera_matrix(K)
    R = RT[:,:3]

    H_rot = np.dot(K, np.dot(R_new, np.dot(R.T, K_inv)))

    T_new = np.dot(R.T, RT[:,3]) - new_location
    T_new = np.dot(R, T_new[:,np.newaxis])
    H_trans = np.array([[1, 0, T_new[0,0]*K[0,0]/distance],
                        [0, 1, T_new[1,0]*K[1,1]/distance],
                        [0, 0,                        1]])

    H = np.dot(H_rot, H_trans)

    return H

def get_largest_contour(img):
    """Return the largest OpenCV contour from a given binary image"""
    c = max(cv.findContours(img,
                            cv.RETR_EXTERNAL,
                            cv.CHAIN_APPROX_NONE)[0],
            key=cv.contourArea)
    return c

def u_to_XY(u_coord, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    XY_transform = distance / K[0,0]
    XY_coord = (u_coord - K[0,2]) * XY_transform
    return XY_coord

def v_to_Z(v_coord, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    Z_transform = distance / K[1,1]
    Z_coord = (K[1,2] - v_coord) * Z_transform
    return Z_coord

def uv_to_XYZ(uv_coord, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    K_inv = inverse_camera_matrix(K)
    R_inv = RT[:,:3].T
    uv_homogenous = np.append(uv_coord, 1)
    XYZ = distance*np.matmul(R_inv, np.matmul(K_inv, uv_homogenous))
    return XYZ

def get_flame_height(img, K, RT, threshold=0.5):
    thresh = np.uint8(img >= threshold)
    flame_contour = get_largest_contour(thresh)
    uv_flame_height = flame_contour[np.argmin(flame_contour[:,0,0]),0,:]
    XYZ_flame_height = uv_to_XYZ(uv_flame_height, K, RT)
    return XYZ_flame_height[2]

def get_contour_top_surface(contour):
    c_sort = contour[np.argsort(np.sum(np.vstack([contour[:,0,0]*100, contour[:,0,1]]).T, axis=1))]
    _, indices = np.unique(c_sort[:,0,0], return_index=True)
    return c_sort[indices,0,:]

def get_contour_bottom_surface(contour):
    c_sort = contour[np.argsort(np.sum(np.vstack([contour[:,0,0]*100, contour[:,0,1]]).T, axis=1))[::-1]]
    _, indices = np.unique(c_sort[:,0,0], return_index=True)
    return c_sort[indices,0,:]

def get_2D_flame_surface(contour, K, RT):
    min_v_with_u = get_contour_top_surface(contour)
    max_v_with_u = get_contour_bottom_surface(contour)

    contour_top_surface = np.array([uv_to_XYZ(uv_coord, K, RT) for uv_coord in min_v_with_u])
    contour_bottom_surface = np.array([uv_to_XYZ(uv_coord, K, RT) for uv_coord in max_v_with_u])

    flame_surface = np.concatenate([contour_bottom_surface[:,:,np.newaxis],
                                    contour_top_surface[:,:,np.newaxis]],
                                    axis=2)
    return flame_surface

def get_volume_from_orthogonal_planes(xz_surface, yz_surface):
    x = np.sort(xz_surface[:,0,0]) * 1000
    y = np.sort(yz_surface[:,1,0]) * 1000
    xi = np.arange(int(np.min(x)), int(np.max(x)))
    yi = np.arange(int(np.min(y)), int(np.max(y)))

    xz_scaled_z_min = xz_surface[:,2,0] * 1000
    yz_scaled_z_min = yz_surface[:,2,0] * 1000

    xz_scaled_z_max = xz_surface[:,2,1] * 1000
    yz_scaled_z_max = yz_surface[:,2,1] * 1000

    xz_z_min = np.interp(xi, x, xz_scaled_z_min)
    yz_z_min = np.interp(yi, y, yz_scaled_z_min)

    xz_z_max = np.interp(xi, x, xz_scaled_z_max)
    yz_z_max = np.interp(yi, y, yz_scaled_z_max)

    z_min = np.max(np.vstack([np.tile(xz_z_min,len(yi)), np.repeat(yz_z_min,len(xi))]).T,axis=1)
    z_max = np.min(np.vstack([np.tile(xz_z_max,len(yi)), np.repeat(yz_z_max,len(xi))]).T,axis=1)

    heights = z_max - z_min
    heights[heights < 0] = 0

    volume = np.sum(heights) * 10**-9

    return volume

def visualise_flame_surface(xz_surface, yz_surface):
    x = np.sort(xz_surface[:,0,0]) * 1000
    y = np.sort(yz_surface[:,1,0]) * 1000
    xi = np.arange(int(np.min(x)), int(np.max(x)))
    yi = np.arange(int(np.min(y)), int(np.max(y)))

    xz_scaled_z_min = xz_surface[:,2,0] * 1000
    yz_scaled_z_min = yz_surface[:,2,0] * 1000

    xz_scaled_z_max = xz_surface[:,2,1] * 1000
    yz_scaled_z_max = yz_surface[:,2,1] * 1000

    xz_z_min = np.interp(xi, x, xz_scaled_z_min)
    yz_z_min = np.interp(yi, y, yz_scaled_z_min)

    xz_z_max = np.interp(xi, x, xz_scaled_z_max)
    yz_z_max = np.interp(yi, y, yz_scaled_z_max)

    z_min = np.max(np.vstack([np.tile(xz_z_min,len(yi)), np.repeat(yz_z_min,len(xi))]).T,axis=1).reshape((len(xi), len(yi)))
    z_max = np.min(np.vstack([np.tile(xz_z_max,len(yi)), np.repeat(yz_z_max,len(xi))]).T,axis=1).reshape((len(xi), len(yi)))

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    xx, yy = np.meshgrid(xi, yi, indexing="ij")
    ax.plot_surface(xx, yy, lower_flame_surface, color="blue")
    ax.plot_surface(xx, yy, upper_flame_surface, color="red")
    ax.legend([mpl.lines.Line2D([0], [0], linestyle="none", color="blue", marker="o"),
               mpl.lines.Line2D([0], [0], linestyle="none", color="red", marker="o")],
              ["Bottom Surface", "Top Surface"])
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    ax.set_xlim3d(-1500, 1500)
    ax.set_ylim3d(-1500, 1500)
    ax.set_zlim3d(-1500, 1500)
    plt.show()
    plt.close()

    return volume

def planar_volume_reconstruction(img_list, K_list, RT_list):
    assert img_list[0].dtype == np.uint8
    assert len(img_list) == 2
    assert len(img_list) == len(K_list) and len(img_list) == len(RT_list)

    contour_list = [get_largest_contour(img) for img in img_list]
    surface_list = [get_2D_flame_surface(contour_list[i], K_list[i], RT_list[i]) for i in range(len(img_list))]
    volume = get_volume_from_orthogonal_planes(surface_list[0], surface_list[1])

    return volume

def get_reconstruction_error(obj_height, obj_width, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    f = (K[0,0] + K[1,1])/2
    transform = distance / f
    err_transform = (distance + obj_width/2) / f

    X = (obj_height / 2) * transform * 2
    Xdx = (obj_height / 2) * err_transform * 2

    diff = abs(Xdx**3 - X**3)

    return diff

def change_camera_perspective(img, K, RT, new_location, new_rotation):
    H = get_perspective_matrix(K, RT, new_location, new_rotation)
    warped = cv.warpPerspective(img, H, dsize=(img.shape[1],img.shape[0]))
    return warped


def main():
    shape_properties = np.load("Cube\\shape-properties.npy")

    img_list = [cv.imread("Cube\\xz-cam.png"), cv.imread("Cube\\yz-cam.png")]
    thresh_list = [cv.threshold(img[:,:,2], 75, 255, cv.THRESH_BINARY)[1] for img in img_list]

    K_list = [np.load("Cube\\xz-cam-K.npy"), np.load("Cube\\yz-cam-K.npy")]
    RT_list = [np.load("Cube\\xz-cam-RT.npy"), np.load("Cube\\yz-cam-RT.npy")]
    assert len(K_list) == len(RT_list) and len(K_list) == len(img_list)

    volume = planar_volume_reconstruction(thresh_list, K_list, RT_list)
    error = volume - shape_properties[0]
    flame_height = max([get_flame_height(thresh_list[i], K_list[i], RT_list[i]) for i in range(len(thresh_list))])
    shape_height = flame_height * (K_list[0][1,1] / 10)

    print("Volume = {} m^3".format(volume))
    print("Error = {} m^3".format(error))
    print("Height = {} m".format(flame_height))
    print("Error = {} m".format(flame_height - shape_properties[4]/2))
    print(get_reconstruction_error(297*2,2,K_list[0], RT_list[0]))
    print(K_list[0])


if __name__ == "__main__":
    main()
