import sys
import os
import numpy as np
import math
from math import radians, degrees

def round_to_zero(array, threshold=1e-6):
    """Return given array with all values below threshold set to zero"""
    return np.where(np.abs(array)>threshold, array, 0)

def eulerAnglesToRotationMatrix(theta):
    """ Convert list of euler angles [X, Y, Z] following the Tait-Bryan
    to a 3x3 rotation matrix, but with a flip in coordinates required
    to match OpenCV.

    Edited from https://learnopencv.com/rotation-matrix-to-euler-angles/"""

    R_x = np.array([[1,         0,                  0                   ],
                    [0,         math.cos(theta[0]), math.sin(theta[0]) ],
                    [0,         -math.sin(theta[0]), math.cos(theta[0])  ]
                    ])

    R_y = np.array([[math.cos(theta[1]),    0,      -math.sin(theta[1])  ],
                    [0,                     1,      0                   ],
                    [math.sin(theta[1]),   0,      math.cos(theta[1])  ]
                    ])

    R_z = np.array([[math.cos(theta[2]),    math.sin(theta[2]),    0],
                    [-math.sin(theta[2]),    math.cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])

    R = np.dot(R_x, np.dot(R_y, R_z))
    # Flip required to match blender angles
    cv_flip = np.array([[1, 0,  0],
                        [0,-1,  0],
                        [0, 0, -1]])
    R = np.dot(cv_flip, R)
    R = round_to_zero(R)

    return R

def get_RT_manually(cam_location, cam_rotation):
    """Return pose matrix from the [X, Y, Z] camera location and
    [theta_x, theta_y, theta_z] euler angles of camera rotation"""
    R = eulerAnglesToRotationMatrix(np.array(cam_rotation))
    tvec = np.array(cam_location)
    T = np.matmul(R, tvec) # convert tvec to camera coordinates
    RT = np.hstack([R, T[:, np.newaxis]])
    return RT

# def get_RT_manually(cam_location, cam_rotation):
#     """Return pose matrix from the [X, Y, Z] camera location and
#     [theta_x, theta_y, theta_z] euler angles of camera rotation"""
#     tvec = np.array(cam_location)
#     R = eulerAnglesToRotationMatrix(np.array(cam_rotation))
#     RT = np.hstack([R, tvec[:, np.newaxis]])
#     return RT

def main():
    # Define varying parameters for the sensitivity study
    default_resolution = np.array([3840, 2160])
    shape_list = ["Cube", "Cone", "Cylinder", "Sphere", "Torus"]
    camera_list = ["xz-cam", "yz-cam"]
    translation_list = [10, 5, 0, -5, -10]
    resolution_list = [default_resolution*n for n in [0.5, 0.75, 1, 1.25, 1.5]]
    offset_list = [n*10**-3 for n in [100, 50, 0, -50, -100]]
    angle_list = [radians(a) for a in [10, 5, 0, -5, -10]]

    for i in range(len(shape_list)):
        for j in range(len(camera_list)):
            for k in range(3):
                for m in range(len(translation_list)):
                    camera_locations = [[0,-20,0], [-20,0,0]]
                    camera_rotations = [[radians(90),0,0], [radians(90),0,radians(270)]]
                    shape = shape_list[i]
                    camera = camera_list[j]
                    camera_location = camera_locations[j]
                    camera_rotation = camera_rotations[j]
                    translation = 0
                    offset = 0
                    angle = 0
                    resolution = default_resolution
                    if k == 0:
                        translation = translation_list[m]
                        resolution = resolution_list[m]
                    elif k == 1:
                        offset = offset_list[m]
                    elif k == 2:
                        angle = angle_list[m]

                    if camera == "xz-cam":
                        camera_location[1] += translation
                        camera_location[0] += offset
                    else:
                        camera_location[0] += translation
                        # camera_location[1] += offset

                    camera_rotation[2] += angle

                    camera_location = tuple(camera_location)
                    camera_rotation = tuple(camera_rotation)

                    camera = camera + "-Z={}-angle={}-offset={}".format(translation, int(degrees(angle)), offset)

                    output_dir = os.path.join(os.getcwd(), shape)
                    if not os.path.isdir(output_dir):
                        os.mkdir(output_dir)

                    RT = get_RT_manually(camera_location, camera_rotation)
                    nRT = np.matrix(RT)
                    filepath = os.path.join(output_dir, f"{camera}-RT.npy")
                    np.save(filepath, nRT)
                    print(filepath)

def distance_analysis():
    default_resolution = np.array([3840, 2160])
    shape_list = ["Cube", "Cone", "Cylinder", "Sphere", "Torus"]
    camera_list = ["xz-cam", "yz-cam"]
    translation_list = [10, 5, 0, -10, -20]
    resolution_list = [default_resolution*n for n in [0.5, 0.75, 1, 1.5, 2]]

    for i in range(len(shape_list)):
        for j in range(len(camera_list)):
            for k in range(len(translation_list)):
                camera_locations = [[0,-20,0], [-20,0,0]]
                camera_rotations = [[radians(90),0,0], [radians(90),0,radians(270)]]
                shape = shape_list[i]
                camera = camera_list[j]
                camera_location = camera_locations[j]
                camera_rotation = camera_rotations[j]
                translation = translation_list[k]
                resolution = resolution_list[k]

                if camera == "xz-cam":
                    camera_location[1] += translation
                else:
                    camera_location[0] += translation

                camera_location = tuple(camera_location)
                camera_rotation = tuple(camera_rotation)

                camera = camera + "-Z={}".format(translation)

                output_dir = os.path.join(os.getcwd(),"distance-analysis", shape)
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)

                RT = get_RT_manually(camera_location, camera_rotation)
                nRT = np.matrix(RT)
                filepath = os.path.join(output_dir, f"{camera}-RT.npy")
                np.save(filepath, nRT)
                print(filepath)

if __name__ == "__main__":
    main()
    distance_analysis()
