import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from scipy.optimize import curve_fit

def fit_func(x, a, b):
    return x*a + b

df = pd.read_csv("blender_scale_analysis.csv", header=0, index_col=0)
df = df.sort_values(by=["shape", "scale"])
df["mm_per_pixel"] = (36/df["scale"]) * (20000/50)

shape_list = ["Cube", "Cone", "Cylinder", "Sphere", "Torus"]
color_list = ["#4477AA", "#EE6677", "#228833", "#CCBB44", "#66CCEE"]
linestyle_list = ["-", "--", "-.", ":", "-"]
marker_list = ["s", "^", "|", "o", "*"]
format_list = ["-s", "--^", "-.|", ":o", "-*"]
shape_heights = [2, 2, 2, 2, 0.5]

fig = plt.figure(figsize=(7.25,4.2))
ax = fig.add_subplot(111)

artist_list = []
for shape, height, c, mk, ls in zip(shape_list, shape_heights, color_list, marker_list, linestyle_list):
    selection = df["shape"] == shape
    baseline = np.logical_and(df["shape"] == shape, df["scale"] == 3840)

    popt, pcov = curve_fit(fit_func, 3840/df[selection]["scale"], df[selection]["volume_error"] - df[baseline]["volume_error"].values)

    ax.scatter(3840/df[selection]["scale"],
               df[selection]["volume_error"] - df[baseline]["volume_error"].values,
               marker=mk,
               linestyle=ls,
               color=c)
    x = np.linspace(1, 10, 100)
    y = fit_func(x, *popt)
    ax.plot(x, y, color=c)
    print(shape)
    print(*popt)


legend_handles = [mlines.Line2D([],
                                [],
                                color=c,
                                marker=m,
                                linestyle=l,
                                label=shape)
                  for c, m, l, shape in zip(color_list,
                                            marker_list,
                                            linestyle_list,
                                            shape_list)]
ax.legend(handles=legend_handles)

ax.set_xlabel(r"1 / $SF$")
ax.set_ylabel("Change in Volume % Error")
ax.set_xlim(0.5, 10.5)
ax.set_xticks(np.linspace(1,10,10))
ax.set_yticks(np.linspace(-5,25,7))
plt.savefig("scale_analysis.png", dpi=300)
plt.savefig("scale_analysis.svg")
plt.savefig("scale_analysis.tif", dpi=1000)
plt.savefig("scale_analysis.eps")
plt.close()


scale_df = pd.read_csv("blender_scale_analysis.csv", header=0, index_col=0)
scale_df = scale_df.sort_values(by=["shape", "scale"])

dist_df = pd.read_csv("blender_distance_analysis.csv", header=0, index_col=0)
dist_df = dist_df.sort_values(by=["shape", "translation"])

fig = plt.figure(figsize=(7.25,4))
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)
ax1 = fig.add_subplot(121)
ax2 = fig.add_subplot(122)

artist_list = []
for shape, height, c, mk, ls in zip(shape_list, shape_heights, color_list, marker_list, linestyle_list):
    selection = dist_df["shape"] == shape
    popt, pcov = curve_fit(fit_func, 1/((dist_df[selection]["translation"]-20) * -1), dist_df[selection]["volume_error"])

    ax1.scatter(1/((dist_df[selection]["translation"]-20) * -1),
                   dist_df[selection]["volume_error"],
                   marker = mk,
                   color=c)
    x = np.linspace(-0.001, 0.1, 100)
    y = fit_func(x, *popt)
    ax1.plot(x, y, color=c)


    selection = scale_df["shape"] == shape
    baseline = np.logical_and(scale_df["shape"] == shape, scale_df["scale"] == 3840)

    popt, pcov = curve_fit(fit_func, 3840/scale_df[selection]["scale"], scale_df[selection]["volume_error"] - scale_df[baseline]["volume_error"].values)

    ax2.scatter(3840/scale_df[selection]["scale"],
               scale_df[selection]["volume_error"] - scale_df[baseline]["volume_error"].values,
               marker=mk,
               linestyle=ls,
               color=c)
    x = np.linspace(0, 10, 100)
    y = fit_func(x, *popt)
    ax2.plot(x, y, color=c)

legend_handles = [mlines.Line2D([],
                                [],
                                color=c,
                                marker=m,
                                linestyle=l,
                                label=shape)
                  for c, m, l, shape in zip(color_list,
                                            marker_list,
                                            linestyle_list,
                                            shape_list)]
ax1.legend(handles=legend_handles)

ax1.set_xlabel(r"1 / $Z_c$ (m$^{-1}$)", fontsize=10)
ax1.set_ylabel("Volume % Error", fontsize=10)
ax1.set_xlim(0,0.105)
ax1.set_ylim(0,150)

ax2.set_xlabel(r"1 / $SF$", fontsize=10)
ax2.set_ylabel("Change in Volume % Error", fontsize=10)
ax2.set_xlim(0,10.5)
# ax.set_xticks(np.linspace(-100,100,5))
ax2.set_yticks(np.linspace(-5,25,7))

plt.savefig("distance_and_scale_combined_plot.png", dpi=300)
plt.savefig("distance_and_scale_combined_plot.svg")
plt.savefig("distance_and_scale_combined_plot.tif", dpi=1000)
plt.savefig("distance_and_scale_combined_plot.eps")
plt.close()
