import numpy as np
import cv2 as cv
import math
import matplotlib.pyplot as plt
import matplotlib as mpl
import scipy.interpolate as interpolate
from math import radians
from scipy.ndimage import zoom

def change_camera_perspective(img, K, RT, new_location, new_rotation):
    H = get_perspective_matrix(K, RT, new_location, new_rotation)
    warped = cv.warpPerspective(img, H, dsize=(img.shape[1],img.shape[0]))
    return warped

def draw_contour(img):
    """Draw largest contour onto binary image"""
    c = get_largest_contour(img)
    cv.drawContours(img, c, -1, (155,0,0), 3)
    return img

def eulerAnglesToRotationMatrix(theta):
    """ Convert list of euler angles [X, Y, Z] following the Tait-Bryan
    to a 3x3 rotation matrix, but with a flip in coordinates required
    to match OpenCV.

    Edited from https://learnopencv.com/rotation-matrix-to-euler-angles/"""

    R_x = np.array([[1,         0,                  0                   ],
                    [0,         math.cos(theta[0]), math.sin(theta[0]) ],
                    [0,         -math.sin(theta[0]), math.cos(theta[0])  ]
                    ])

    R_y = np.array([[math.cos(theta[1]),    0,      -math.sin(theta[1])  ],
                    [0,                     1,      0                   ],
                    [math.sin(theta[1]),   0,      math.cos(theta[1])  ]
                    ])

    R_z = np.array([[math.cos(theta[2]),    math.sin(theta[2]),    0],
                    [-math.sin(theta[2]),    math.cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])

    R = np.dot(R_x, np.dot(R_y, R_z))
    # Flip required to match blender angles
    cv_flip = np.array([[1, 0,  0],
                        [0,-1,  0],
                        [0, 0, -1]])
    R = np.dot(cv_flip, R)
    R = round_to_zero(R)

    return R

def get_2D_flame_surface(contour, K, RT, scale=1):
    min_v_with_u = get_contour_top_surface(contour)
    max_v_with_u = get_contour_bottom_surface(contour)

    contour_top_surface = np.array([uv_to_XYZ(uv_coord, K, RT, scale) for uv_coord in min_v_with_u])
    contour_bottom_surface = np.array([uv_to_XYZ(uv_coord, K, RT, scale) for uv_coord in max_v_with_u])

    flame_surface = np.concatenate([contour_bottom_surface[:,:,np.newaxis],
                                    contour_top_surface[:,:,np.newaxis]],
                                    axis=2)

    return flame_surface

def get_contour_bottom_surface(contour):
    c_sort = contour[np.argsort(np.sum(np.vstack([contour[:,0,0]*100, contour[:,0,1]]).T, axis=1))[::-1]]
    _, indices = np.unique(c_sort[:,0,0], return_index=True)
    return c_sort[indices,0,:]

def get_contour_top_surface(contour):
    c_sort = contour[np.argsort(np.sum(np.vstack([contour[:,0,0]*100, contour[:,0,1]]).T, axis=1))]
    _, indices = np.unique(c_sort[:,0,0], return_index=True)
    return c_sort[indices,0,:]

def get_flame_height(img, K, RT, threshold=0.5, scale=1):
    thresh = np.uint8(img >= threshold)
    flame_contour = get_largest_contour(thresh)
    if flame_contour is None:
        flame_height = 0
    else:
        surface = get_2D_flame_surface(flame_contour, K, RT, scale)
        max_height = np.max(surface[:,2,1])
        min_height = np.min(surface[:,2,0])
        flame_height = max_height - min_height
    return flame_height

def get_flame_tip_location(img, K, RT, threshold=0.5, scale=1):
    thresh = np.uint8(img >= threshold)
    flame_contour = get_largest_contour(thresh)
    if flame_contour is None:
        max_height = 0
    else:
        surface = get_2D_flame_surface(flame_contour, K, RT, scale)
        max_height = np.max(surface[:,2,1])
    return max_height

def get_largest_contour(img):
    """Return the largest OpenCV contour from a given binary image"""
    contours = cv.findContours(img,
                            cv.RETR_EXTERNAL,
                            cv.CHAIN_APPROX_NONE)[0]
    if len(contours) > 0:
        c = max(cv.findContours(img,
                                cv.RETR_EXTERNAL,
                                cv.CHAIN_APPROX_NONE)[0],
                key=cv.contourArea)
    else:
        c = None
    return c

def get_perspective_matrix(
    K,
    RT,
    new_location=(-5*np.sqrt(2),-5*np.sqrt(2),-0.2),
    new_rotation=(90,0,-55)):
    """Return the homography that transforms an image from the current camera
    to one of the same scene taken by the same camera at a new location and
    rotation, assuming the same projective distance from the scene"""
    theta = [radians(angle) for angle in new_rotation]
    R_new = eulerAnglesToRotationMatrix(theta)

    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    K_inv = inverse_camera_matrix(K)
    R = RT[:,:3]

    H_rot = np.dot(K, np.dot(R_new, np.dot(R.T, K_inv)))

    T_new = np.dot(R.T, RT[:,3]) - new_location
    T_new = np.dot(R, T_new[:,np.newaxis])
    H_trans = np.array([[1, 0, T_new[0,0]*K[0,0]/distance],
                        [0, 1, T_new[1,0]*K[1,1]/distance],
                        [0, 0,                        1]])

    H = np.dot(H_rot, H_trans)

    return H

def get_volume_from_orthogonal_planes(xz_surface, yz_surface, K_list, RT_list, scale=1, units_scaling=1):
    distance_list = [np.sqrt(np.dot(RT[:,3], RT[:,3])) for RT in RT_list]
    x_mm_per_px = (distance_list[0] / K_list[0][0,0]) * units_scaling / scale
    y_mm_per_px = (distance_list[1] / K_list[1][0,0]) * units_scaling / scale
    # print(f"mm per pixel = X:{x_mm_per_px}, Y:{y_mm_per_px}")

    x = np.sort(xz_surface[:,0,0]) * units_scaling
    y = np.sort(yz_surface[:,1,0]) * units_scaling

    s = 4
    nx = math.ceil((np.max(x) - np.min(x)) * s / x_mm_per_px)
    ny = math.ceil((np.max(y) - np.min(y)) * s / y_mm_per_px)

    xi = np.linspace(np.min(x), np.max(x), nx)
    yi = np.linspace(np.min(y), np.max(y), ny)

    xz_scaled_z_min = xz_surface[:,2,0] * units_scaling
    yz_scaled_z_min = yz_surface[:,2,0] * units_scaling

    xz_scaled_z_max = xz_surface[:,2,1] * units_scaling
    yz_scaled_z_max = yz_surface[:,2,1] * units_scaling

    min_surface_len = min([len(xz_scaled_z_min),
                           len(xz_scaled_z_max),
                           len(yz_scaled_z_min),
                           len(yz_scaled_z_max)])

    if min_surface_len <= 4:
        volume = 0
    else:
        xz_z_min_curve = interpolate.UnivariateSpline(x, xz_scaled_z_min, ext=3)
        xz_z_max_curve = interpolate.UnivariateSpline(x, xz_scaled_z_max, ext=3)
        yz_z_min_curve = interpolate.UnivariateSpline(y, yz_scaled_z_min, ext=3)
        yz_z_max_curve = interpolate.UnivariateSpline(y, yz_scaled_z_max, ext=3)

        # Enforce that maximum height from both views be equal (which should be true in idealised case)
        max_diff = np.max(xz_scaled_z_max) - np.max(yz_scaled_z_max)
        min_diff = np.min(xz_scaled_z_min) - np.min(yz_scaled_z_min)

        xx, yy = np.meshgrid(xi, yi)
        z_min = np.max(np.stack([xz_z_min_curve(xx), yz_z_min_curve(yy) + min_diff]), axis=0)
        z_max = np.min(np.stack([xz_z_max_curve(xx), yz_z_max_curve(yy) + max_diff]), axis=0)

        heights = z_max - z_min
        heights[heights < 0] = 0

        volume = np.sum(heights) * x_mm_per_px * y_mm_per_px * s**-2 * 10**-9

    return volume

def inverse_camera_matrix(K):
    """Return inverted OpenCV intrinsic camera matrix"""
    f_x = K[0,0]
    f_y = K[1,1]
    c_x = K[0,2]
    c_y = K[1,2]
    alpha = 1/(f_x*f_y)

    K_inv = np.float32([
        [f_y,   0, -f_y * c_x],
        [  0, f_x, -f_x * c_y],
        [  0,   0,    f_x*f_y]
        ]) * alpha

    return K_inv

def isRotationMatrix(R) :
    """Check if a numpy array is a valid rotation matrix"""
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype = R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6

def planar_volume_reconstruction(img_list, K_list, RT_list, scale=1, units_scaling=1):
    assert img_list[0].dtype == np.uint8
    assert len(img_list) == 2
    assert len(img_list) == len(K_list) and len(img_list) == len(RT_list)

    contour_list = [get_largest_contour(img) for img in img_list]
    if any([contour is None for contour in contour_list]) or any([len(contour) < 4 for contour in contour_list]):
        volume = 0
    else:
        surface_list = [get_2D_flame_surface(contour_list[i], K_list[i], RT_list[i], scale) for i in range(len(img_list))]
        volume = get_volume_from_orthogonal_planes(surface_list[0], surface_list[1], K_list, RT_list, scale, units_scaling)

    return volume

def rotationMatrixToEulerAngles(R) :
    """Convert a 3x3 OpenCV rotation matrix to the Euler angles used by
    Blender"""

    assert(isRotationMatrix(R))

    sy = math.sqrt(R[0,0] * R[0,0] +  R[1,0] * R[1,0])

    singular = sy < 1e-6

    if  not singular :
        x = math.atan2(R[2,1] , R[2,2])
        z = math.atan2(-R[2,0], sy)
        y = math.atan2(R[1,0], R[0,0])
    else :
        x = math.atan2(-R[1,2], R[1,1])
        z = math.atan2(-R[2,0], sy)
        y = 0

    angles = np.array([x, y, z])
    angles = round_to_zero(angles)

    return np.array([x, y, z])

def round_to_zero(array, threshold=1e-6):
    """Return given array with all values below threshold set to zero"""
    return np.where(np.abs(array)>threshold, array, 0)

def threshold(img, boundary=0.8):
    hist = cv.calcHist([img], [0], None, [56], [200, 256], accumulate=False)
    total_px = np.sum(hist.flatten())
    thresh_limit = 254 - np.nonzero((np.cumsum(hist[::-1]) - total_px*boundary)>=0)[0][0]
    ret, thresh = cv.threshold(img, thresh_limit, 255, cv.THRESH_BINARY)
    return thresh

def u_to_XY(u_coord, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    XY_transform = distance / K[0,0]
    XY_coord = (u_coord - K[0,2]) * XY_transform
    return XY_coord

def uv_to_XYZ(uv_coord, K, RT, scale=1):
    uv_coord = uv_coord/scale
    Z_c = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    # Z_c = RT[2,3]
    K_inv = inverse_camera_matrix(K)
    R_inv = RT[:,:3].T
    tvec = RT[:,3]
    uv_homogenous = np.append(uv_coord, 1)
    XYZ = np.matmul(R_inv, np.matmul(K_inv, Z_c * uv_homogenous) - tvec)
    return XYZ

def v_to_Z(v_coord, K, RT):
    distance = np.sqrt(np.dot(RT[:,3], RT[:,3]))
    Z_transform = distance / K[1,1]
    Z_coord = (K[1,2] - v_coord) * Z_transform
    return Z_coord

def visualise_flame_surface(xz_surface, yz_surface, K_list, RT_list, filename, scale=1, units_scaling=1):
    distance_list = [np.sqrt(np.dot(RT[:,3], RT[:,3])) for RT in RT_list]
    x_mm_per_px = (distance_list[0] / K_list[0][0,0]) * units_scaling / scale
    y_mm_per_px = (distance_list[1] / K_list[1][0,0]) * units_scaling / scale

    x = np.sort(xz_surface[:,0,0]) * units_scaling
    y = np.sort(yz_surface[:,1,0]) * units_scaling

    s = 4
    nx = math.ceil((np.max(x) - np.min(x)) * s / x_mm_per_px)
    ny = math.ceil((np.max(y) - np.min(y)) * s / y_mm_per_px)

    xi = np.linspace(np.min(x), np.max(x), nx)
    yi = np.linspace(np.min(y), np.max(y), ny)

    xz_scaled_z_min = xz_surface[:,2,0] * units_scaling
    yz_scaled_z_min = yz_surface[:,2,0] * units_scaling

    xz_scaled_z_max = xz_surface[:,2,1] * units_scaling
    yz_scaled_z_max = yz_surface[:,2,1] * units_scaling

    min_surface_len = min([len(xz_scaled_z_min),
                           len(xz_scaled_z_max),
                           len(yz_scaled_z_min),
                           len(yz_scaled_z_max)])

    if min_surface_len > 4:
        xz_z_min_curve = interpolate.UnivariateSpline(x, xz_scaled_z_min, ext=3)
        xz_z_max_curve = interpolate.UnivariateSpline(x, xz_scaled_z_max, ext=3)
        yz_z_min_curve = interpolate.UnivariateSpline(y, yz_scaled_z_min, ext=3)
        yz_z_max_curve = interpolate.UnivariateSpline(y, yz_scaled_z_max, ext=3)

        # Enforce that maximum height from both views be equal (which should be true in idealised case)
        max_diff = np.max(xz_scaled_z_max) - np.max(yz_scaled_z_max)
        min_diff = np.min(xz_scaled_z_min) - np.min(yz_scaled_z_min)

        xx, yy = np.meshgrid(xi, yi)
        z_min = np.max(np.stack([xz_z_min_curve(xx), yz_z_min_curve(yy) + min_diff]), axis=0)
        z_max = np.min(np.stack([xz_z_max_curve(xx), yz_z_max_curve(yy) + max_diff]), axis=0)

        fig = plt.figure()
        ax = fig.gca(projection='3d')
        ax.plot_surface(xx[10:-10,10:-10], yy[10:-10,10:-10], z_min[10:-10,10:-10], color="blue", alpha=0.8)
        ax.plot_surface(xx[10:-10,10:-10], yy[10:-10,10:-10], z_max[10:-10,10:-10], color="red", alpha=0.8)
        # ax.plot3D(xi, -500*np.ones(len(xi)), xz_z_max_curve(xi), lw=1.5, color="black", zorder=10)
        # ax.plot3D(1000*np.ones(len(yi)), yi, yz_z_max_curve(yi) + max_diff, lw=1.5, ls="dashed", color="black", zorder=10)
        ax.legend([mpl.lines.Line2D([0], [0], linestyle="none", color="blue", marker="o"),
                   mpl.lines.Line2D([0], [0], linestyle="none", color="red", marker="o")],
                  ["Bottom Surface", "Top Surface"])

        # Plot wall
        x = np.linspace(-1000,1000,10)
        z = np.linspace(0,2500,10)
        xx, zz = np.meshgrid(x, z)
        yy = np.full(xx.shape, 330)
        ax.plot_surface(xx,yy,zz,lw=1, color="brown", alpha=0.5, zorder=-2)
        # Plot burner
        x = np.linspace(-800,800,10)
        y = np.linspace(-80,80,10)
        xx, yy = np.meshgrid(x, y)
        zz = np.full(xx.shape, 0)
        ax.plot_surface(xx,yy,zz,lw=1, color="black", alpha=0.5, zorder=-2)

        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_xlim3d(-1500, 1500)
        ax.set_ylim3d(-1650, 350)
        ax.set_zlim3d(-150, 3000)
        plt.savefig(filename, dpi=300)
        plt.close()

        # z_max_smooth = interpolate.RectBivariateSpline(xi, yi, z_max)
        # z_min_smooth = interpolate.RectBivariateSpline(xi, yi, z_min)
        # fig = plt.figure()
        # ax = fig.gca(projection='3d')
        # nx = 2000
        # ny = 2000
        # xi = np.linspace(np.min(x), np.max(x), nx)
        # yi = np.linspace(np.min(y), np.max(y), ny)
        # xx, yy = np.meshgrid(xi, yi, indexing="ij")
        # ax.plot_surface(xx, yy, z_min_smooth(xi, yi), color="blue")
        # ax.plot_surface(xx, yy, z_max_smooth(xi, yi), color="red")
        # ax.legend([mpl.lines.Line2D([0], [0], linestyle="none", color="blue", marker="o"),
        #            mpl.lines.Line2D([0], [0], linestyle="none", color="red", marker="o")],
        #           ["Bottom Surface", "Top Surface"])
        # ax.set_xlabel("X")
        # ax.set_ylabel("Y")
        # ax.set_zlabel("Z")
        # ax.set_xlim3d(-1500, 1500)
        # ax.set_ylim3d(-1500, 1500)
        # ax.set_zlim3d(-1500, 1500)
        # plt.savefig(f"smooth_{filename}", dpi=300)
        # plt.close()

        # plt.figure()
        # plt.imshow(z_max, interpolation="gaussian")
        # plt.savefig(f"contour_{filename}", dpi=300)
        # plt.close()

    return True

def visualise_shape_surface(xz_surface, yz_surface, K_list, RT_list, filename, scale=1, units_scaling=1):
    distance_list = [np.sqrt(np.dot(RT[:,3], RT[:,3])) for RT in RT_list]
    x_mm_per_px = (distance_list[0] / K_list[0][0,0]) * units_scaling / scale
    y_mm_per_px = (distance_list[1] / K_list[1][0,0]) * units_scaling / scale

    x = np.sort(xz_surface[:,0,0]) * units_scaling
    y = np.sort(yz_surface[:,1,0]) * units_scaling

    s = 4
    nx = math.ceil((np.max(x) - np.min(x)) * s / x_mm_per_px)
    ny = math.ceil((np.max(y) - np.min(y)) * s / y_mm_per_px)

    xi = np.linspace(np.min(x), np.max(x), nx)
    yi = np.linspace(np.min(y), np.max(y), ny)

    xz_scaled_z_min = xz_surface[:,2,0] * units_scaling
    yz_scaled_z_min = yz_surface[:,2,0] * units_scaling

    xz_scaled_z_max = xz_surface[:,2,1] * units_scaling
    yz_scaled_z_max = yz_surface[:,2,1] * units_scaling

    xz_z_min_curve = interpolate.UnivariateSpline(x, xz_scaled_z_min)
    xz_z_max_curve = interpolate.UnivariateSpline(x, xz_scaled_z_max)
    yz_z_min_curve = interpolate.UnivariateSpline(y, yz_scaled_z_min)
    yz_z_max_curve = interpolate.UnivariateSpline(y, yz_scaled_z_max)

    # Enforce that maximum height from both views be equal (which should be true in idealised case)
    max_diff = np.max(xz_scaled_z_max) - np.max(yz_scaled_z_max)
    min_diff = np.min(xz_scaled_z_min) - np.min(yz_scaled_z_min)

    xx, yy = np.meshgrid(xi, yi)
    z_min = np.max(np.stack([xz_z_min_curve(xx), yz_z_min_curve(yy) + min_diff]), axis=0)
    z_max = np.min(np.stack([xz_z_max_curve(xx), yz_z_max_curve(yy) + max_diff]), axis=0)

    fig = plt.figure(figsize=(4.0, 2.2))
    ax = fig.gca(projection='3d')
    ax.plot_surface(xx, yy, z_min, color="blue")
    ax.plot_surface(xx, yy, z_max, color="red")
    ax.legend([mpl.lines.Line2D([0], [0], linestyle="none", color="blue", marker="o"),
               mpl.lines.Line2D([0], [0], linestyle="none", color="red", marker="o")],
              ["Bottom Surface", "Top Surface"], fontsize=8)


    ax.set_xlabel("X", fontsize=10)
    ax.set_ylabel("Y", fontsize=10)
    ax.set_zlabel("Z", fontsize=10)
    ax.set_xlim3d(-1.1, 1.1)
    ax.set_ylim3d(-1.1, 1.1)
    ax.set_zlim3d(-1.1, 1.1)
    ax.set_xticks(np.linspace(-1,1,5))
    ax.set_yticks(np.linspace(-1,1,5))
    ax.set_zticks(np.linspace(-1,1,5))
    plt.rc('xtick', labelsize=8)
    plt.rc('ytick', labelsize=8)
    plt.savefig(filename+".svg")
    plt.savefig(filename+".png", dpi=300)
    plt.close()

    return True

def save_flame_surface_img(img_list, K_list, RT_list, filename, scale=1):
    assert img_list[0].dtype == np.uint8
    assert len(img_list) == 2
    assert len(img_list) == len(K_list) and len(img_list) == len(RT_list)

    contour_list = [get_largest_contour(img) for img in img_list]
    surface_list = [get_2D_flame_surface(contour_list[i], K_list[i], RT_list[i], scale) for i in range(len(img_list))]
    visualise_flame_surface(surface_list[0], surface_list[1], K_list, RT_list, filename, scale=scale)

    return True

def save_shape_surface_img(img_list, K_list, RT_list, filename, scale=1):
    assert img_list[0].dtype == np.uint8
    assert len(img_list) == 2
    assert len(img_list) == len(K_list) and len(img_list) == len(RT_list)

    contour_list = [get_largest_contour(img) for img in img_list]
    surface_list = [get_2D_flame_surface(contour_list[i], K_list[i], RT_list[i], scale) for i in range(len(img_list))]
    visualise_shape_surface(surface_list[0], surface_list[1], K_list, RT_list, filename, scale=scale)

    return True

def get_RT_manually(cam_location, cam_rotation):
    """Return pose matrix from the [X, Y, Z] camera location and
    [theta_x, theta_y, theta_z] euler angles of camera rotation"""
    tvec = np.array(cam_location)
    R = eulerAnglesToRotationMatrix(np.array(cam_rotation))
    RT = np.hstack([R, tvec[:, np.newaxis]])
    return RT

if __name__ == "__main__":
    cam_location = [1,1,1]
    cam_rotation = [radians(90), 0, 0]

    print(get_RT_manually(cam_location, cam_rotation))

    x = np.linspace(-800,800,10)
    y = np.linspace(-80,80,10)
    xx, yy = np.meshgrid(x, y)
    zz = np.full(xx.shape, 0)
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.plot_surface(xx,yy,zz,lw=1, color="black")
    plt.show()
    plt.close()
